theory Embedding_Trs
imports
  Trs
  "$AFP/Well_Quasi_Orders/Kruskal"
  "$AFP/Well_Quasi_Orders/Kruskal_Terms"
begin

text {*Terms constructed using @{const Fun} only, are finite trees.*}
interpretation term_tree: finite_tree Fun "fst \<circ> the \<circ> root" args
  by (unfold_locales) simp_all
text {*The set of \emph{ground trees} (i.e., terms without considering arities
of function symbols).*}
inductive_set gtrees for A where
  "f \<in> A \<Longrightarrow> \<forall>t\<in>set ts. t \<in> gtrees A \<Longrightarrow> Fun f ts \<in> gtrees A"

text {*The set @{term "finite_tree.trees Fun A"} is the same
as the set of ground-terms over @{term A}.*}
lemma gtrees_simp [simp]:
  "gtrees A = term_tree.trees A"
  (is "?lhs = ?rhs")
proof
  show "?lhs \<subseteq> ?rhs"
  proof
    fix t assume "t \<in> ?lhs" thus "t \<in> ?rhs"
      by (induct) (rule term_tree.trees_intros, auto)
  qed
next
  show "?rhs \<subseteq> ?lhs"
  proof
    fix t assume "t \<in> ?rhs" thus "t \<in> ?lhs"
      by (induct) (auto intro: gtrees.intros)
  qed
qed

definition Emb where
  "Emb \<equiv> { (Fun f ts, t) | f ts t. t \<in> set ts }"

definition Hembeq where
  "Hembeq P \<equiv> Emb \<union>
    { (Fun f ss, Fun g ts) | f ss g ts. P\<^sup>=\<^sup>= g f \<and> sublisteq ts ss }"

lemma list_hembeq_mono:
  assumes "\<And>s t. P s t \<longrightarrow> Q s t"
  shows "list_hembeq P s t \<longrightarrow> list_hembeq Q s t"
proof
  assume "list_hembeq P s t" thus "list_hembeq Q s t"
    by (induct) (insert assms, auto)
qed

lemma reflclp_mono:
  assumes "\<And>s t. P s t \<longrightarrow> Q s t"
  shows "P\<^sup>=\<^sup>= s t \<longrightarrow> Q\<^sup>=\<^sup>= s t"
  using assms by auto

abbreviation hembeq where "hembeq P \<equiv> term_tree.tree_hembeq P"

text {*Rewriting step without substitution.*}
inductive_set cstep for R :: "('a, 'b) trs" where
  base [intro]: "(l, r) \<in> R \<Longrightarrow> (l, r) \<in> cstep R" |
  step [intro]: "(s, t) \<in> cstep R \<Longrightarrow> (C\<langle>s\<rangle>, C \<langle>t\<rangle>) \<in> cstep R"

lemma cstepD:
  assumes "(s, t) \<in> cstep R"
  shows "\<exists>l r C. (l, r) \<in> R \<and> s = C\<langle>l\<rangle> \<and> t = C\<langle>r\<rangle>"
using assms
proof (induct)
  case (base l r)
  hence "(l, r) \<in> R \<and> l = \<box>\<langle>l\<rangle> \<and> r = \<box>\<langle>r\<rangle>" by simp
  thus ?case by blast
next
  case (step s t C)
  then obtain l r D where "(l, r) \<in> R"
    and s: "s = D\<langle>l\<rangle>" and t: "t = D \<langle>r\<rangle>" by auto
  moreover have "C\<langle>s\<rangle> = (C \<circ>c D)\<langle>l\<rangle>" and "C\<langle>t\<rangle> = (C \<circ>c D)\<langle>r\<rangle>" by (simp_all add: s t)
  ultimately show ?case by blast
qed

lemma cstepE [elim]:
  assumes "(s, t) \<in> cstep R"
    and "\<And>C l r. (l, r) \<in> R \<Longrightarrow> s = C\<langle>l\<rangle> \<Longrightarrow> t = C\<langle>r\<rangle> \<Longrightarrow> P"
  shows P
  using assms and cstepD by blast

lemma hembeq_ctxt:
  assumes "hembeq P s t"
  shows "hembeq P (C\<langle>s\<rangle>) (C\<langle>t\<rangle>)"
  using assms by (induct C) auto

lemma list_hembeq_cstep_HembeqD:
  assumes "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hembeq P))\<^sup>*) xs zs"
    (is "list_hembeq (\<lambda>s t. (t, s) \<in> ?R) xs zs" is "list_hembeq ?P xs zs")
  shows "\<exists>ys. sublisteq ys zs \<and> length ys = length xs \<and>
    (\<forall>i<length ys. (ys ! i, xs ! i) \<in> (cstep (Hembeq P))\<^sup>*)"
using assms
proof (induct)
  case (list_hembeq_Nil ys)
  thus ?case by auto
next
  case (list_hembeq_Cons xs zs z)
  then obtain ys where list_hembeq: "sublisteq ys zs" and "length ys = length xs"
    and "\<forall>i<length ys. (ys ! i, xs ! i) \<in> ?R" by auto
  moreover have "sublisteq ys (z # zs)" using list_hembeq by auto
  ultimately show ?case by blast
next
  case (list_hembeq_Cons2 x z xs zs)
  then obtain ys where list_hembeq: "sublisteq ys zs"
    and len: "length ys = length xs"
    and *: "\<forall>i<length ys. (ys ! i, xs ! i) \<in> ?R" by auto
  let ?ys = "z # ys" and ?xs = "x # xs"
  from list_hembeq have "sublisteq ?ys (z # zs)" by auto
  moreover have "length ?ys = length ?xs" using len by auto
  moreover have "\<forall>i<length ?ys. (?ys ! i, ?xs ! i) \<in> ?R"
  proof (intro allI impI)
    fix i
    assume i: "i < length ?ys"
    show "(?ys ! i, ?xs ! i) \<in> ?R"
      using i and * and `(\<lambda>s t. (t, s) \<in> ?R)\<^sup>=\<^sup>= x z`
      by (cases i) auto
  qed
  ultimately show ?case by blast
qed

lemma args_cstep_Hembeq:
  assumes "length ss = length ts"
    and args: "\<forall>i<length ss. (ss ! i, ts ! i) \<in> (cstep (Hembeq P))\<^sup>*"
  shows "(Fun f ss, Fun f ts) \<in> (cstep (Hembeq P))\<^sup>*"
proof -
  have "ctxt_closed (cstep (Hembeq P))" by (rule) auto
  from args_steps_imp_steps [OF this assms(1)] and assms(2)
    show ?thesis  by metis
qed

lemma Fun_cstep_Hembeq:
  assumes "P\<^sup>=\<^sup>= g f" and "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hembeq P))\<^sup>*) ts ss"
  shows "(Fun f ss, Fun g ts) \<in> (cstep (Hembeq P))\<^sup>+"
proof -
  from list_hembeq_cstep_HembeqD [OF assms(2)] and args_cstep_Hembeq
    obtain us where emb: "sublisteq us ss"
    and "(Fun g us, Fun g ts) \<in> (cstep (Hembeq P))\<^sup>*" by force
  moreover have "(Fun f ss, Fun g us) \<in> (cstep (Hembeq P))\<^sup>+"
  proof -
    from emb and `P\<^sup>=\<^sup>= g f` have "(Fun f ss, Fun g us) \<in> (cstep (Hembeq P))"
      by (auto simp: Hembeq_def Emb_def)
    thus ?thesis by auto
  qed
  ultimately show ?thesis by auto
qed

lemma csteps_ctxt:
  assumes "(s, t) \<in> (cstep R)\<^sup>+"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> (cstep R)\<^sup>+"
  using assms by (induct) (auto intro: trancl_trans)

lemma hembeq_imp_cstep_Hembeq_trancl:
  assumes "hembeq P s t"
  shows "(t, s) \<in> (cstep (Hembeq P))\<^sup>+"
using assms
proof (induct)
  case (tree_hembeq_base t ts f)
  hence "(Fun f ts, t) \<in> Hembeq P" by (auto simp: Emb_def Hembeq_def)
  thus ?case by auto
next
  case (tree_hembeq_sublisteq f g ss ts)
  from this(2) show ?case
  proof (induct)
    case (list_hembeq_Nil ys)
    have "sublisteq [] ys" by auto
    with `P\<^sup>=\<^sup>= f g` have "(Fun g ys, Fun f []) \<in> Hembeq P" by (auto simp: Hembeq_def)
    thus ?case by auto 
  next
    case (list_hembeq_Cons xs ys y)
    have "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hembeq P))\<^sup>*) xs ys"
      by (rule list_hembeq_mono [rule_format]) (insert list_hembeq_Cons, auto)
    hence "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hembeq P))\<^sup>*) xs (y # ys)" by auto
    from Fun_cstep_Hembeq [of P f g, OF `P\<^sup>=\<^sup>= f g` this] show ?case by simp
  next
    case (list_hembeq_Cons2 x y xs ys)
    have "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hembeq P))\<^sup>*) xs ys"
      by (rule list_hembeq_mono [rule_format]) (insert list_hembeq_Cons2, auto)
    with list_hembeq_Cons2
      have "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hembeq P))\<^sup>*) (x # xs) (y # ys)" by auto
    from Fun_cstep_Hembeq [of P f g, OF `P\<^sup>=\<^sup>= f g` this] show ?case by simp
  qed
next
  case (tree_hembeq_trans s t u) thus ?case by auto
next
  case (tree_hembeq_ctxt s t f ss1 ss2)
  from csteps_ctxt [OF this(2), of "More f ss1 \<box> ss2"]
    show ?case by simp
qed

lemma cstep_Hembeq_imp_hemb:
  assumes "(s, t) \<in> cstep (Hembeq P)"
  shows "hembeq P t s"
using assms
proof
  fix C l r
  assume "(l, r) \<in> Hembeq P" and s: "s = C\<langle>l\<rangle>" and t: "t = C\<langle>r\<rangle>"
  thus ?thesis
  proof (unfold Hembeq_def, elim UnE)
    assume "(l, r) \<in> Emb"
    then obtain f ts where l: "l = Fun f ts" and "r \<in> set ts" by (auto simp: Emb_def)
    hence "hembeq P r l" unfolding l by auto
    from hembeq_ctxt [OF this] show ?thesis by (simp add: s t)
  next
    assume "(l, r) \<in> {(Fun f ss, Fun g ts) | f ss g ts. P\<^sup>=\<^sup>= g f \<and> sublisteq ts ss}"
    then obtain f ss g ts where l: "l = Fun f ss" and r: "r = Fun g ts"
      and "sublisteq ts ss" and "P\<^sup>=\<^sup>= g f" by auto
    from `sublisteq ts ss` and s and t and l and r show ?thesis
    proof (induct arbitrary: l r)
      case (list_hembeq_Nil ss)
      have "list_hembeq (hembeq P) [] ss" by auto
      with `P\<^sup>=\<^sup>= g f` have "hembeq P r l" unfolding list_hembeq_Nil by auto
      from hembeq_ctxt [OF this] show ?case unfolding list_hembeq_Nil by blast
    next
      case (list_hembeq_Cons ts ss u)
      moreover have "sublisteq ss (u # ss)" by auto
      ultimately have *: "sublisteq ts (u # ss)" by auto
      from term_tree.tree_hembeq_sublisteq [OF `P\<^sup>=\<^sup>= g f` this]
        have "hembeq P r l"
        unfolding list_hembeq_Cons .
      from hembeq_ctxt [OF this] show ?case unfolding list_hembeq_Cons by blast
    next
      case (list_hembeq_Cons2 u v ts ss)
      hence *: "sublisteq (u # ts) (v # ss)" by auto
      hence "hembeq P r l" unfolding list_hembeq_Cons2 using `P\<^sup>=\<^sup>= g f` by auto
      from hembeq_ctxt [OF this] show ?case unfolding list_hembeq_Cons2 by blast
    qed
  qed
qed

lemma cstep_Hembeq_trancl_imp_hemb:
  assumes "(s, t) \<in> (cstep (Hembeq P))\<^sup>+"
  shows "hembeq P t s"
using assms
proof (induct s t)
  case (r_into_trancl s t)
  from this show ?case by (rule cstep_Hembeq_imp_hemb)
next
  case (trancl_into_trancl s t u)
  from `(t, u) \<in> cstep (Hembeq P)` have "hembeq P u t" by (rule cstep_Hembeq_imp_hemb)
  with trancl_into_trancl show ?case by blast
qed

lemma list_hembeq_eq_map:
  assumes "sublisteq ss ts"
  shows "sublisteq (map f ss) (map f ts)"
  using assms by (induct) auto

lemma Hembeq_subst:
  assumes "(l, r) \<in> Hembeq P"
  shows "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> Hembeq P"
    (is "(?l, ?r) \<in> ?R")
using assms [unfolded Hembeq_def]
proof
  assume "(l, r) \<in> Emb"
  then obtain f ts where l: "l = Fun f ts" and "r \<in> set ts" by (auto simp: Emb_def)
  let ?ts = "map (\<lambda>t. t \<cdot> \<sigma>) ts"
  have "?l = Fun f ?ts" unfolding l by simp
  moreover have "?r \<in> set ?ts" using `r \<in> set ts` by auto
  ultimately show "(?l, ?r) \<in> ?R" by (auto simp: Hembeq_def Emb_def)
next
  assume "(l, r) \<in> {(Fun f ss, Fun g ts) | f ss g ts. P\<^sup>=\<^sup>= g f \<and> sublisteq ts ss}"
  then obtain f g ss ts where "P\<^sup>=\<^sup>= g f" and l: "l = Fun f ss" and r: "r = Fun g ts"
    and emb: "sublisteq ts ss" by auto
  let ?ts = "map (\<lambda>t. t \<cdot> \<sigma>) ts"
  let ?ss = "map (\<lambda>t. t \<cdot> \<sigma>) ss"
  let ?l = "l \<cdot> \<sigma>"
  let ?r = "r \<cdot> \<sigma>"
  from list_hembeq_eq_map [OF emb] have "sublisteq ?ts ?ss" by auto
  with `P\<^sup>=\<^sup>= g f` show "(?l, ?r) \<in> ?R"
    unfolding Hembeq_def l r by auto
qed

lemma cstep_Hembeq_eq_rstep_Hembeq:
  "cstep (Hembeq P) = rstep (Hembeq P)" (is "?L = ?R")
proof
  show "?R \<subseteq> ?L"
  proof
    fix s t
    assume "(s, t) \<in> ?R"
    thus "(s, t) \<in> ?L"
    proof
      fix l r C \<sigma>
      assume "(l, r) \<in> Hembeq P" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      from Hembeq_subst [OF this(1)] have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> cstep (Hembeq P)" by auto
      from cstep.step [OF this] show ?thesis unfolding s t by blast
    qed
  qed
next
  show "?L \<subseteq> ?R"
  proof
    fix s t
    assume "(s, t) \<in> ?L"
    thus "(s, t) \<in> ?R"
      by (induct s t) auto
  qed
qed

lemma hembeq_eq_rstep_Hembeq_trancl:
  "hembeq P s t \<longleftrightarrow> (t, s) \<in> (rstep (Hembeq P))\<^sup>+"
  unfolding cstep_Hembeq_eq_rstep_Hembeq [symmetric]
  using hembeq_imp_cstep_Hembeq_trancl
    and cstep_Hembeq_trancl_imp_hemb by blast

lemma wqo_on_hemb:
  assumes "wqo_on P A"
  shows "wqo_on (hembeq P) (gtrees A)"
  using term_tree.wqo_on_trees [OF assms] by simp

lemmas wfp_on_hemb = wqo_on_hemb [THEN wqo_on_imp_wfp_on]

inductive_set gterms for F where
  "(f, n) \<in> F \<Longrightarrow> length ts = n \<Longrightarrow> \<forall>t\<in>set ts. t \<in> gterms F \<Longrightarrow> Fun f ts \<in> gterms F"

text {*The set @{term "finite_tree.terms Fun F"} is the same
as the set of ground-terms over @{term F}.*}
lemma gterms_simp [simp]:
  "gterms F = term_tree.terms F"
  (is "?lhs = ?rhs")
proof
  show "?lhs \<subseteq> ?rhs"
  proof
    fix t assume "t \<in> ?lhs" thus "t \<in> ?rhs"
      by (induct) (rule term_tree.terms.intros, auto)
  qed
next
  show "?rhs \<subseteq> ?lhs"
  proof
    fix t assume "t \<in> ?rhs" thus "t \<in> ?lhs"
      by (induct) (auto intro: gterms.intros)
  qed
qed

definition Hemb where
  "Hemb F P \<equiv> Emb \<union>
    { (Fun f ss, Fun g ts) | f ss g ts.
    (g, length ts) \<in> F \<and> (f, length ss) \<in> F \<and>
    P (g, length ts) (f, length ss) \<and> sublisteq ts ss }"

abbreviation hemb where "hemb \<equiv> term_tree.term_hemb"

lemma ground_funas_term_subseteq_gterms_conv:
  "ground t \<and> funas_term t \<subseteq> F \<longleftrightarrow> t \<in> gterms F"
proof
  assume "ground t \<and> funas_term t \<subseteq> F"
  then show "t \<in> gterms F"
    by (induct t) (auto, rule term_tree.terms.intros, auto)
next
  assume "t \<in> gterms F"
  then show "ground t \<and> funas_term t \<subseteq> F"
    by (induct) auto
qed

lemma hemb_ctxt:
  assumes "funas_ctxt C \<subseteq> F" and "ground_ctxt C" and "hemb F P s t"
  shows "hemb F P (C\<langle>s\<rangle>) (C\<langle>t\<rangle>)"
  using assms
  by (induct C)
     (auto intro!: term_tree.term_hemb_ctxt,
       (metis SUP_le_iff ground_funas_term_subseteq_gterms_conv gterms_simp)+)

lemma list_hembeq_cstep_HembD:
  assumes "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hemb F P))\<^sup>*) xs zs"
    (is "list_hembeq (\<lambda>s t. (t, s) \<in> ?R) xs zs" is "list_hembeq ?P xs zs")
  shows "\<exists>ys. sublisteq ys zs \<and> length ys = length xs \<and>
    (\<forall>i<length ys. (ys ! i, xs ! i) \<in> (cstep (Hemb F P))\<^sup>*)"
using assms
proof (induct)
  case (list_hembeq_Nil ys)
  thus ?case by auto
next
  case (list_hembeq_Cons xs zs z)
  then obtain ys where list_hembeq: "sublisteq ys zs" and "length ys = length xs"
    and "\<forall>i<length ys. (ys ! i, xs ! i) \<in> ?R" by auto
  moreover have "sublisteq ys (z # zs)" using list_hembeq by auto
  ultimately show ?case by blast
next
  case (list_hembeq_Cons2 x z xs zs)
  then obtain ys where list_hembeq: "sublisteq ys zs"
    and len: "length ys = length xs"
    and *: "\<forall>i<length ys. (ys ! i, xs ! i) \<in> ?R" by auto
  let ?ys = "z # ys" and ?xs = "x # xs"
  from list_hembeq have "sublisteq ?ys (z # zs)" by auto
  moreover have "length ?ys = length ?xs" using len by auto
  moreover have "\<forall>i<length ?ys. (?ys ! i, ?xs ! i) \<in> ?R"
  proof (intro allI impI)
    fix i
    assume i: "i < length ?ys"
    show "(?ys ! i, ?xs ! i) \<in> ?R"
      using i and * and `(\<lambda>s t. (t, s) \<in> ?R)\<^sup>=\<^sup>= x z`
      by (cases i) auto
  qed
  ultimately show ?case by blast
qed

lemma args_cstep_Hemb:
  assumes "length ss = length ts"
    and args: "\<forall>i<length ss. (ss ! i, ts ! i) \<in> (cstep (Hemb F P))\<^sup>*"
  shows "(Fun f ss, Fun f ts) \<in> (cstep (Hemb F P))\<^sup>*"
proof -
  have "ctxt_closed (cstep (Hemb F P))" by (rule) auto
  from args_steps_imp_steps [OF this assms(1)] and assms(2)
    show ?thesis  by metis
qed

lemma hemb_imp_cstep_Hemb_trancl:
  assumes "hemb F P s t"
  shows "(t, s) \<in> (cstep (Hemb F P))\<^sup>+"
using assms
proof (induct)
  case (term_hemb_base f n ts t)
  hence "(Fun f ts, t) \<in> Hemb F P" by (auto simp: Emb_def Hemb_def)
  thus ?case by auto
next
  case (term_hemb_sublisteq f n g m ss ts)
  then have "(Fun g ts, Fun f ss) \<in> Hemb F P"
    unfolding Hemb_def by auto
  then show ?case by auto
next
  case (term_hemb_trans s t u) thus ?case by auto
next
  case (term_hemb_ctxt s t f n ss1 ss2)
  from csteps_ctxt [OF this(2), of "More f ss1 \<box> ss2"]
    show ?case by simp
qed

lemma ctxt_gterms:
  assumes "C\<langle>t\<rangle> \<in> gterms F" shows "t \<in> gterms F"
  using assms by (induct C) (auto elim: term_tree.terms.cases)

lemma gtermsD:
  "t \<in> gterms F \<Longrightarrow> ground t \<and> funas_term t \<subseteq> F"
  unfolding ground_funas_term_subseteq_gterms_conv [symmetric] by blast

lemma ctxt_gtermsD:
  assumes "C\<langle>t\<rangle> \<in> gterms F"
  shows "ground_ctxt C \<and> ground t \<and> funas_ctxt C \<subseteq> F \<and> funas_term t \<subseteq> F"
proof -
  from gtermsD [OF assms] have "ground (C\<langle>t\<rangle>)" and "funas_term (C\<langle>t\<rangle>) \<subseteq> F" by auto
  then show ?thesis by auto
qed

lemma Hemb_imp_hemb:
  assumes "(l, r) \<in> Hemb F P" and "l \<in> gterms F" and "r \<in> gterms F"
  shows "hemb F P r l"
proof -
  { assume "(l, r) \<in> Emb"
    then obtain f ts where l: "l = Fun f ts" and "r \<in> set ts" unfolding Emb_def by auto
    have "hemb F P r l" unfolding l
    proof (rule term_tree.term_hemb_base)
      from `l \<in> gterms F` show "(f, length ts) \<in> F"
        and "\<forall>t\<in>set ts. t \<in> term_tree.terms F" unfolding l by (cases, auto)+
      show "r \<in> set ts" by fact
    qed simp }
  moreover
  { fix f ss and g ts
    assume l: "l = Fun f ss" and r: "r = Fun g ts"
      and "P (g, length ts) (f, length ss)"
      and "sublisteq ts ss"
    have "hemb F P r l" unfolding l r
    proof (rule term_tree.term_hemb_sublisteq)
      from `l \<in> gterms F` show "(f, length ss) \<in> F"
        and "\<forall>t\<in>set ss. t \<in> term_tree.terms F" unfolding l by (cases, auto)+
      from `r \<in> gterms F` show "(g, length ts) \<in> F"
        and "\<forall>t\<in>set ts. t \<in> term_tree.terms F" unfolding r by (cases, auto)+
      show "P (g, length ts) (f, length ss)" by fact
      show "sublisteq ts ss" by fact
    qed simp+ }
  ultimately show ?thesis using `(l, r) \<in> Hemb F P` unfolding Hemb_def by blast
qed

lemma cstep_Hemb_imp_hemb':
  assumes "(s, t) \<in> cstep (Hemb F P)" and *: "s \<in> gterms F" and **: "t \<in> gterms F"
  shows "hemb F P t s"
using assms(1)
proof
  fix C l r
  assume "(l, r) \<in> Hemb F P" and s: "s = C\<langle>l\<rangle>" and t: "t = C\<langle>r\<rangle>"
  from ctxt_gterms [OF * [unfolded s]] have "l \<in> gterms F" .
  moreover
  from ctxt_gterms [OF ** [unfolded t]] have "r \<in> gterms F" .
  ultimately
  have "hemb F P r l"
    using Hemb_imp_hemb [OF `(l, r) \<in> Hemb F P`] by blast
  with * and ** show ?thesis unfolding s t
  proof (induct C)
    case Hole then show ?case by simp
  next
    case (More f ss1 D ss2)
    let ?C = "More f ss1 \<box> ss2"
    from More have "?C\<langle>D\<langle>l\<rangle>\<rangle> \<in> gterms F"
      and "?C\<langle>D\<langle>r\<rangle>\<rangle> \<in> gterms F" by auto
    with ctxt_gterms have "D\<langle>l\<rangle> \<in> gterms F"
      and "D\<langle>r\<rangle> \<in> gterms F" by blast+
    with More have "hemb F P (D\<langle>r\<rangle>) (D\<langle>l\<rangle>)" by blast
    have "hemb F P (Fun f (ss1 @ D\<langle>r\<rangle> # ss2)) (Fun f (ss1 @ D\<langle>l\<rangle> # ss2))"
    proof
      show "hemb F P (D\<langle>r\<rangle>) (D\<langle>l\<rangle>)" by fact
      from `?C\<langle>D\<langle>l\<rangle>\<rangle> \<in> gterms F` have "Fun f (ss1 @ D\<langle>l\<rangle> # ss2) \<in> gterms F" by simp
      then show "(f, Suc (length ss1 + length ss2)) \<in> F"
        and "\<forall>t\<in> set (ss1@ss2). t \<in> term_tree.terms F" by (cases, auto)+
    qed simp+
    then show ?case by simp
  qed
qed

lemma restrict_cstep_Hemb_imp_hemb:
  assumes "(s, t) \<in> restrict (cstep (Hemb F P)) (gterms F)"
  shows "hemb F P t s"
  using assms
  by (simp add: restrict_def cstep_Hemb_imp_hemb')

lemma Hemb_subst:
  assumes "(l, r) \<in> Hemb F P"
  shows "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> Hemb F P"
    (is "(?l, ?r) \<in> ?R")
using assms [unfolded Hemb_def]
proof
  assume "(l, r) \<in> Emb"
  then obtain f ts where l: "l = Fun f ts" and "r \<in> set ts" by (auto simp: Emb_def)
  let ?ts = "map (\<lambda>t. t \<cdot> \<sigma>) ts"
  have "?l = Fun f ?ts" unfolding l by simp
  moreover have "?r \<in> set ?ts" using `r \<in> set ts` by auto
  ultimately show "(?l, ?r) \<in> ?R" by (auto simp: Hemb_def Emb_def)
next
  assume "(l, r) \<in> {(Fun f ss, Fun g ts) | f ss g ts. (g, length ts) \<in> F \<and>
    (f, length ss) \<in> F \<and> P (g, length ts) (f, length ss) \<and> sublisteq ts ss}"
  then obtain f g ss ts where "(g, length ts) \<in> F" and "(f, length ss) \<in> F"
    and "P (g, length ts) (f, length ss)" and l: "l = Fun f ss" and r: "r = Fun g ts"
    and emb: "sublisteq ts ss" by auto
  let ?ts = "map (\<lambda>t. t \<cdot> \<sigma>) ts"
  let ?ss = "map (\<lambda>t. t \<cdot> \<sigma>) ss"
  let ?l = "l \<cdot> \<sigma>"
  let ?r = "r \<cdot> \<sigma>"
  from list_hembeq_eq_map [OF emb] have "sublisteq ?ts ?ss" by auto
  with `(g, length ts) \<in> F` and `(f, length ss) \<in> F` and`P (g, length ts) (f, length ss)`
    show "(?l, ?r) \<in> ?R"
    unfolding Hemb_def l r by auto
qed

lemma cstep_Hemb_rstep_Hemb_conv:
  "cstep (Hemb F P) = rstep (Hemb F P)" (is "?L = ?R")
proof
  show "?R \<subseteq> ?L"
  proof
    fix s t
    assume "(s, t) \<in> ?R"
    thus "(s, t) \<in> ?L"
    proof
      fix l r C \<sigma>
      assume "(l, r) \<in> Hemb F P" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      from Hemb_subst [OF this(1)] have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> cstep (Hemb F P)" by auto
      from cstep.step [OF this] show ?thesis unfolding s t by blast
    qed
  qed
next
  show "?L \<subseteq> ?R"
  proof
    fix s t
    assume "(s, t) \<in> ?L"
    thus "(s, t) \<in> ?R"
      by (induct s t) auto
  qed
qed

lemma sig_step_cstep_conv:
  "(s, t) \<in> sig_step F (cstep R) \<and> ground s \<and> ground t \<longleftrightarrow>
    (s, t) \<in> cstep R \<and> s \<in> gterms F \<and> t \<in> gterms F"
  unfolding sig_step_def
  unfolding ground_funas_term_subseteq_gterms_conv [symmetric]
  by blast

lemma Hemb_preserves_gterms:
  assumes "(l, r) \<in> Hemb F P" and "l \<in> gterms F"
  shows "r \<in> gterms F"
  using assms
proof -
  { assume "(l, r) \<in> Emb"
    then obtain f ts where l: "l = Fun f ts" and "r \<in> set ts" unfolding Emb_def by auto
    from `l \<in> gterms F` have "\<forall>t\<in>set ts. t \<in> gterms F"
      unfolding l by (cases) auto
    then have "r \<in> gterms F" using `r \<in> set ts` by blast }
  moreover
  { fix f ss and g ts
    assume l: "l = Fun f ss" and r: "r = Fun g ts"
      and "(g, length ts) \<in> F" and "(f, length ss) \<in> F"
      and "P (g, length ts) (f, length ss)"
      and "sublisteq ts ss"
   from `l \<in> gterms F` have "\<forall>t\<in>set ss. t \<in> gterms F"
     unfolding l by (cases) auto
   with `sublisteq ts ss` have "\<forall>t\<in>set ts. t \<in> gterms F"
     by (induct) auto
   with `(g, length ts) \<in> F` have "r \<in> gterms F"
     unfolding r by auto }
  ultimately show ?thesis using assms(1) unfolding Hemb_def by blast
qed

lemma cstep_Hemb_preserves_gterms:
  assumes "(s, t) \<in> cstep (Hemb F P)" and "s \<in> gterms F"
  shows "t \<in> gterms F"
using `(s, t) \<in> cstep (Hemb F P)`
proof
  fix C l r
  assume "(l, r) \<in> Hemb F P" and s: "s = C\<langle>l\<rangle>" and t: "t = C\<langle>r\<rangle>"
  from `s \<in> gterms F` have "ground_ctxt C" and "ground l"
    and "funas_ctxt C \<subseteq> F" and "funas_term l \<subseteq> F"
    unfolding s
    by (blast dest: ctxt_gtermsD)+
  then have "l \<in> gterms F"
    unfolding ground_funas_term_subseteq_gterms_conv [symmetric] by blast
  with `(l, r) \<in> Hemb F P` have "r \<in> gterms F" by (rule Hemb_preserves_gterms)
  then have "ground r" and "funas_term r \<subseteq> F"
    unfolding ground_funas_term_subseteq_gterms_conv [symmetric] by blast+
  with `ground_ctxt C` and `funas_ctxt C \<subseteq> F`
    have "ground t" and "funas_term t \<subseteq> F"
    unfolding t by auto
  then show "t \<in> gterms F"
    unfolding ground_funas_term_subseteq_gterms_conv [symmetric] by blast
qed

lemma cstep_Hemb_imp_hemb:
  assumes "(s, t) \<in> cstep (Hemb F P)" and "s \<in> gterms F"
  shows "hemb F P t s"
  using assms cstep_Hemb_imp_hemb' and cstep_Hemb_preserves_gterms by blast

lemma cstep_Hemb_trancl_preserves_gterms:
  assumes "(s, t) \<in> (cstep (Hemb F P))\<^sup>+" and "s \<in> gterms F"
  shows "t \<in> gterms F"
  using assms
  by (induct rule: trancl.induct)
     (auto dest: cstep_Hemb_preserves_gterms)

lemma cstep_Hemb_trancl_imp_hemb:
  assumes "(s, t) \<in> (cstep (Hemb F P))\<^sup>+" and "s \<in> gterms F"
  shows "hemb F P t s"
  using assms
  by (induct rule: trancl.induct)
     (blast dest: cstep_Hemb_trancl_preserves_gterms cstep_Hemb_imp_hemb)+

text {*The inductive definition of strict homeomorphic embedding on (ground) terms
is equivalent to the formulation as transitive closure of the rewrite relation of an
embedding TRS (as used in the term rewriting literature).*}
lemma hemb_rstep_Hemb_conv:
  "hemb F P t s \<longleftrightarrow> s \<in> gterms F \<and> (s, t) \<in> (rstep (Hemb F P))\<^sup>+"
  (is "?hemb \<longleftrightarrow> ?Hemb")
proof
  assume ?Hemb then show ?hemb
    unfolding cstep_Hemb_rstep_Hemb_conv [symmetric]
    by (blast dest: cstep_Hemb_trancl_imp_hemb)
next
  assume ?hemb
  then show ?Hemb
    unfolding cstep_Hemb_rstep_Hemb_conv [symmetric]
    by (auto dest: hemb_imp_cstep_Hemb_trancl term_tree.term_hemb_imp_terms)
qed

(*
lemma wpo_on_hemb:
  assumes "wpo_on P F"
  shows "wpo_on (hemb F P) (gterms F)"
  using term_tree.wpo_on_terms [OF assms] by simp

lemmas wfp_on_hemb = wpo_on_hemb [THEN wpo_on_imp_wfp_on]
*)

end

