(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Multiset_Extension_bonus
imports Multiset2 Multiset_Extension2

begin

(** This theory contains all the lemmas that are not
needed in IsaFor but that were proven anyway. *)

text {* Alternative definition that is equivalent if ns = s \<or> eq. *}

definition s_mul_ext2 :: "'a relation \<Rightarrow> 'a relation \<Rightarrow> ('a multiset) relation"
 where "s_mul_ext2 ns s A B \<equiv> \<exists> as bs A' B'. (
  A = multiset_of as + A' \<and>
  B = multiset_of bs + B' \<and>
  length as = length bs \<and>
  A' \<noteq> {#} \<and>
  (\<forall>i. i < length bs \<longrightarrow> (as ! i) = (bs ! i)) \<and>
  (\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> s a b))
  )"

text{* The two definitions are equivalent *}

lemma s_mul_ext_eq:
  fixes A :: "'a multiset"
  assumes "locally_reflexive ns (A + B)" (is "?lr (A + B)")
  and eq_dec_sound: "\<And>a b. ((eq_dec :: 'a \<Rightarrow> 'a \<Rightarrow> bool) a b = (a = b))"
  and "locally_splittable s ns eq (A + B)" (is "?lsp (A + B)")
  shows "(s_mul_ext ns s A B) \<equiv> (s_mul_ext2 ns s A B)"
proof-
 {
   assume "s_mul_ext2 ns s A B"
   from this obtain as bs A' B' where
     A_part: "A = multiset_of as + A'" and
     B_part: "B = multiset_of bs + B'" and
     aslbs: "length as = length bs" and
     A'_ne: "A' \<noteq> {#}" and
     eq_as_bs: "(\<forall>i. i < length bs \<longrightarrow> (as ! i) = (bs ! i))" and
     s_A'_B': "(\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> s a b))" by auto
     {
       fix i assume i_bd: "i < length bs"
       from nth_mem_multiset_of[OF i_bd] and B_part
        have bs_in_B: "(bs ! i) \<in># B" by simp
       from `?lr (A + B)`[unfolded locally_reflexive_def, rule_format]
        and bs_in_B and eq_as_bs and i_bd
        have "ns (as ! i) (bs ! i)" by simp
     }
     hence "\<forall>i. i < length bs \<longrightarrow> ns (as ! i) (bs ! i)" by simp
     with A_part and B_part and aslbs and A'_ne and s_A'_B'
      have "s_mul_ext ns s A B" unfolding s_mul_ext_def by metis
 }
 hence C1: "s_mul_ext2 ns s A B \<Longrightarrow> s_mul_ext ns s A B" by assumption
 {
   assume sAB: "s_mul_ext ns s A B"
   from this obtain as bs A' B' where
     A_part: "A = multiset_of as + A'" and
     B_part: "B = multiset_of bs + B'" and
     aslbs: "length as = length bs" and
     A'_ne: "A' \<noteq> {#}" and
     ns_as_bs: "(\<forall>i. i < length bs \<longrightarrow> ns (as ! i) (bs ! i))" and
     s_A'_B': "(\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> s a b))" by auto
   let ?cs = "fst (fst (partition2 eq_dec as bs))"
   let ?ds = "snd (fst (partition2 eq_dec as bs))"
   let ?es = "fst (snd (partition2 eq_dec as bs))"
   let ?fs = "snd (snd (partition2 eq_dec as bs))"
    {
      fix i
      assume i_bd: "i < length ?ds"
      with filter2_length[of eq_dec as bs]
       have i_bd2: "i < length ?cs" unfolding partition2_def by simp
      from i_bd2 and partition2_sound_P[rule_format]
        and eq_dec_sound[of "?cs ! i"]
      have "(?cs ! i) = (?ds ! i)" unfolding partition2_def
       by metis
   }
   hence eq_cs_ds: "\<forall>i. i < length ?ds \<longrightarrow> (?cs ! i) = (?ds ! i)" by simp
   {
     fix i
     assume i_bd: "i < length ?fs"
     with filter2_length[of _ as bs]
       have i_bd2: "i < length ?es" unfolding partition2_def by simp
      from i_bd2 and partition2_sound_nP[rule_format]
        and eq_dec_sound[of "?es ! i"]
      have "(?es ! i) \<noteq> (?fs ! i)" unfolding partition2_def
       by metis
   }


oops


text{* The non-strict order is antisymmetric*}

lemma ns_mul_ext_antisym_aux :
  assumes "locally_antisymmetric ns (A + B)"   (is "?las (A + B)")
  and "locally_included ns s (A + B)"          (is "?lin (A + B)")
  and "locally_irreflexive s (A + B)"          (is "?li (A + B)")
  and "locally_transitive s (A + B)"           (is "?lt_s (A + B)")
  and "locally_transitive ns (A + B)"          (is "?lt_ns (A + B)")
  and "locally_compatible_l s ns (A + B)"      (is "?lcl (A + B)")
  and "locally_compatible_r ns s (A + B)"      (is "?lcr (A + B)")
  and "locally_splittable ns s (op =) (A + B)" (is "?lsp (A + B)")
  shows "size A = n \<longrightarrow> ns_mul_ext ns s A B \<longrightarrow> ns_mul_ext ns s B A \<longrightarrow> A = B"
using assms
proof (induct n arbitrary: A B)
 case 0
(* If A is empty then either B is empty or B \<le># A cannot hold *)
  {
   assume sA_null: "size A = 0"
     and AnsB: "ns_mul_ext ns s A B"
     and s_ns: "?lin (A + B)"
   with size_eq_0_iff_empty have A_empty: "A = {#}" by simp
   have "A = B"
    proof (cases B)
     case 1
      with A_empty show ?thesis by simp
     next
     case (2 B' b)
      hence b_in_B: "b \<in># B" by auto
      from ` ?lin (A + B)` and AnsB and b_in_B and ns_mul_ext_domination[of _ _ A B]
       obtain a where
       a_in_A: "a \<in># A" and
       nsab: "ns a b" by metis
      from a_in_A and A_empty show ?thesis by simp
    qed
  }
  thus ?case using 0 by metis
 next
(* If A is not empty then we can remove a local maxima which is
     in A & B *)
 case (Suc n)
  note IH = Suc(1)
  {
    assume sAsn: "size A = Suc n"
      and "?lt_s (A + B)"
      and "?li (A + B)"
      and AnsB: "ns_mul_ext ns s A B"
      and BnsA: "ns_mul_ext ns s B A"
    from `?lt_s (A + B)` and lt_trans_l have s_trans: "?lt_s A" by fast
    from `?li (A + B)` and li_trans_l have s_irrefl: "?li A" by fast
    from sAsn have "A \<noteq> {#}" by auto
    with s_trans  and s_irrefl and local_maxima_multiset obtain atop where
      atop_in_A: "atop \<in># A" and
      atop_top: "\<forall>a. a \<in># A \<longrightarrow> \<not> s a atop" by metis
    from BnsA obtain bs as B' A' where
      B_part: "B = multiset_of bs + B'" and
      A_part: "A = multiset_of as + A'" and
      aslbs: "length as = length bs" and
      ns_bs_as: "\<forall>i. i < length as \<longrightarrow> ns (bs ! i) (as ! i)" and
      s_B'_A': "\<forall>a. a \<in># A' \<longrightarrow> (\<exists>b. b \<in># B' \<and> s b a)" by auto
     {
       assume a_in_A': "atop \<in># A'"
       with s_B'_A' obtain b where b_in_B': "b \<in># B'" and sbatop: "s b atop" by auto
       with B_part have b_in_B: "b \<in># B" by auto
       from `?lin (A + B)` and b_in_B and AnsB and ns_mul_ext_domination
        obtain "a'" where
        a'_in_A: "a' \<in># A" and
        nsa'b: "ns a' b" by blast
       from sbatop and nsa'b and atop_in_A and b_in_B and a'_in_A
        and `?lcr (A + B)`[unfolded locally_compatible_r_def, rule_format]
        have "s a' atop" by auto
       with atop_top and a'_in_A have "False" by simp
     }
    hence C1: "atop \<in># A' \<Longrightarrow> ?case" by auto
    {
      assume a_in_as: "atop \<in># multiset_of as"
       and "?lin (A + B)"
       and "?lt_ns (A + B)"
      with in_multiset_of_idx[of as atop] obtain i where
        i_bd: "i < length as" and
        a_eq: "atop = as ! i" by auto
      with ns_bs_as have nsbatop: "ns (bs ! i) atop" by auto
      from i_bd and aslbs and nth_mem_multiset_of
       have "(bs ! i) \<in># multiset_of bs" by auto
       with B_part have bs_in_B: "(bs ! i) \<in># B" by auto
      with `?lin (A + B)` and AnsB and ns_mul_ext_domination obtain "a'" where
        a'_in_A: "a' \<in># A" and
        nsa'b: "ns a' (bs ! i)" by blast
      with nsbatop and bs_in_B and atop_in_A
        and `?lt_ns (A + B)`[unfolded locally_transitive_def, rule_format]
        have nsa'atop: "ns a' atop" by auto
      (** We now know that atop \<le> a'. We distinguish the two following cases:
        -- either atop < a' (SC1)
        -- or atop = a' (SC2) *)
       {
         assume "s a' atop"
         with atop_top and a'_in_A have "False" by simp
       }
      hence SC1: "s a' atop \<Longrightarrow> ?case" by simp
       {
         assume aeqa': "atop = a'"
           and ns_antisym: "?las (A + B)"
          from aeqa' and nsa'b and nsbatop and atop_in_A and bs_in_B
            and ns_antisym[unfolded locally_antisymmetric_def, rule_format]
           have b_eq: "atop = bs ! i" by simp
          from i_bd and aslbs and nth_mem_multiset_of
           have "bs ! i \<in># multiset_of bs" by simp
          with b_eq[symmetric] and B_part have atop_in_B: "atop \<in># B" by auto
          (* On prend Ai = "multiset_of (remove_nth i as) + A'"
          Bi = "multiset_of (remove_nth i bs) + B'" *)
           (** Ai a la bonne taille pour l'appel recursif *)
         have Ai: "A - {#atop#} = multiset_of (remove_nth i as) + A'"
           proof-
            from a_eq and diff_union_swap_multiset[symmetric] and A_part
             and nth_mem_multiset_of[OF i_bd]
             have "A - {#atop#} = multiset_of as - {#as ! i#} + A'" by simp
            with remove_nth_soundness[OF i_bd, symmetric]
             show ?thesis by simp
           qed
         have Bi: "B - {#atop#} = multiset_of (remove_nth i bs) + B'"
           proof-
            from aslbs and i_bd have i_bd2: "i < length bs" by simp
            from b_eq and diff_union_swap_multiset[symmetric] and B_part
             and nth_mem_multiset_of[OF i_bd2]
             have "B - {#atop#} = multiset_of bs - {#bs ! i#} + B'" by simp
            with remove_nth_soundness[OF i_bd2, symmetric]
             show ?thesis by simp
           qed
          from remove_nth_length[OF i_bd] and size_multiset_of[of "remove_nth i as"]
            and size_multiset_of[of as, symmetric]
           have "size (multiset_of (remove_nth i as)) = size (multiset_of as) - 1" by simp
           with size_union
            have sAi: "size (multiset_of (remove_nth i as) + A')
              = size (multiset_of as) - 1 + size A'" by simp

(** FIXME: sledgehammer was involved... *)

           hence "size (multiset_of (remove_nth i as) + A')
              = size (multiset_of as) + size A' - 1" by (metis Suc_eq_plus1 `length as = size (multiset_of as)` diff_add_assoc gr0I gr_implies_not0 i_bd less_eq_Suc_le_raw nat_add_commute plus_nat.add_0 union_commute)

           with A_part[symmetric] and size_union[symmetric]
            have "size (multiset_of (remove_nth i as) + A') = size A - 1" by auto
           with sAsn and Ai have sAin: "size (A - {#atop#}) = n" by simp
             (** Bi \<ge># Ai *)
           from aslbs have aislbis: "length (remove_nth i as) = length (remove_nth i bs)"
             by auto
           from aslbs and ns_bs_as and remove_nth_P_compat[of _ _ ns]
            have ns_bis_ais: "\<forall>j. j < length (remove_nth i as) \<longrightarrow>
              ns ((remove_nth i bs) ! j) ((remove_nth i as) ! j)" by simp
           from Ai and Bi and aislbis[symmetric] and aslbs and ns_bis_ais and s_B'_A'
            have  ns_Bi_Ai:"ns_mul_ext ns s (B - {#atop#}) (A - {#atop#})"
            unfolding ns_mul_ext_def by blast

            (** Ai \<ge># Bi *)
           from AnsB obtain ds cs D' C' where
             B_part2: "B = multiset_of cs + C'" and
             A_part2: "A = multiset_of ds + D'" and
             cslds: "length cs = length ds" and
             ns_ds_cs: "\<forall>i. i < length cs \<longrightarrow> ns (ds ! i) (cs ! i)" and
             s_D'_C': "\<forall>c. c \<in># C' \<longrightarrow> (\<exists>d. d \<in># D' \<and> s d c)" by auto
           have atop_in_cs: "atop \<in># multiset_of cs"
           proof (cases "atop \<in># multiset_of cs")
            case True thus ?thesis by assumption
            next
            case False
             with B_part2 and atop_in_B have "atop \<in># C'" by auto
             with s_D'_C' obtain d where
               d_in_D': "d \<in># D'" and
               s_d_atop: "s d atop" by auto
             with A_part2 have "d \<in># A" by simp
             with s_d_atop and atop_top show ?thesis by simp
           qed
          from in_multiset_of_idx[OF atop_in_cs] obtain i where
            i_bd: "i < length cs" and
            c_eq: "atop = cs ! i" by auto
          have Ci: "B - {#atop#} = multiset_of (remove_nth i cs) + C'"
           proof-
            from c_eq and diff_union_swap_multiset[symmetric] and B_part2 and atop_in_cs
             have "B - {#atop#} = multiset_of cs - {#cs ! i#} + C'" by simp
            with remove_nth_soundness[OF i_bd, symmetric]
             show ?thesis by simp
           qed
         have d_eq: "atop = ds ! i"
           proof (cases "atop = ds ! i")
             case True thus ?thesis by assumption
             next
             case False note H = this
              from i_bd and nth_mem_multiset_of[of i ds] and cslds and A_part2
               have ds_in_A: "ds ! i \<in># A" by auto
              from ns_ds_cs and c_eq and i_bd have "ns (ds ! i) atop" by auto
              with atop_in_A and ds_in_A
               `?lsp (A + B)`[unfolded locally_splittable_def, rule_format]
               and H have Hf: "s (ds ! i) atop" by auto
              from i_bd and cslds and nth_mem_multiset_of
               have "ds ! i \<in># multiset_of ds" by auto
              with A_part2 have atop_in_A: "ds ! i \<in># A" by auto
              with atop_top and Hf show ?thesis by auto
           qed
         have Di: "A - {#atop#} = multiset_of (remove_nth i ds) + D'"
           proof-
            from cslds and i_bd have i_bd2: "i < length ds" by simp
            from d_eq and diff_union_swap_multiset[symmetric] and A_part2
             and nth_mem_multiset_of[OF i_bd2]
             have "A - {#atop#} = multiset_of ds - {#ds ! i#} + D'" by simp
            with remove_nth_soundness[OF i_bd2, symmetric]
             show ?thesis by simp
           qed
           from cslds have cisldis: "length (remove_nth i cs) = length (remove_nth i ds)"
             by auto
           from cslds and ns_ds_cs and remove_nth_P_compat[of _ _ ns]
            have ns_dis_cis: "\<forall>j. j < length (remove_nth i cs) \<longrightarrow>
              ns ((remove_nth i ds) ! j) ((remove_nth i cs) ! j)" by simp
           from Ci and Di and cisldis[symmetric] and ns_dis_cis and s_D'_C'
            have  ns_Ai_Bi:"ns_mul_ext ns s (A - {#atop#}) (B - {#atop#})"
            unfolding ns_mul_ext_def by blast

           from atop_in_A and atop_in_B
            have Heq: "(A - {#atop#}) + (B - {#atop#}) = A + B - ({#atop#} + {#atop#})"
            by (metis a'_in_A aeqa' diff_add diff_union_single_conv union_commute)
             (* sledgehammer involved. simplify? *)
          from Heq and `?las (A + B)`and las_minus
           have las: "?las ((A - {#atop#}) + (B - {#atop#}))"
           by simp
          from Heq and `?lin (A + B)`and lin_minus
           have lin: "?lin ((A - {#atop#}) + (B - {#atop#}))"
           by simp
          from Heq and `?li (A + B)`and li_minus
           have li: "?li ((A - {#atop#}) + (B - {#atop#}))"
           by simp
          from Heq and `?lt_s (A + B)`and lt_minus
           have lt_s: "?lt_s ((A - {#atop#}) + (B - {#atop#}))"
           by simp
          from Heq and `?lt_ns (A + B)`and lt_minus
           have lt_ns: "?lt_ns ((A - {#atop#}) + (B - {#atop#}))"
           by simp
          from Heq and `?lcl (A + B)`and lcl_minus
           have lcl: "?lcl ((A - {#atop#}) + (B - {#atop#}))"
           by simp
          from Heq and `?lcr (A + B)`and lcr_minus
           have lcr: "?lcr ((A - {#atop#}) + (B - {#atop#}))"
           by simp
          from Heq and `?lsp (A + B)`and lsp_minus
           have lsp: "?lsp ((A - {#atop#}) + (B - {#atop#}))"
           by simp
          from sAin and ns_Ai_Bi and ns_Bi_Ai and IH
            and las and lin and li and lt_s and lt_ns and lcl and lcr and lsp
            have "A - {#atop#} = B - {#atop#}" by blast
           hence "A - {#atop#} + {#atop#} = B - {#atop#} + {#atop#}" by simp
           with atop_in_A and atop_in_B and insert_DiffM2 have ?case by auto
       }
      hence SC2: "atop = a' \<Longrightarrow> ?case" using Suc by metis
      from SC1 and SC2 and nsa'atop and atop_in_A and a'_in_A
       and `?lsp (A + B)`[unfolded locally_splittable_def, rule_format]
       have "?case" by auto
    }
    hence C2: "atop \<in># multiset_of as \<Longrightarrow> ?case" using Suc by metis
    from C1 and C2 and A_part and atop_in_A have ?case by auto
  }
  thus ?case using Suc by fast
qed

lemma ns_mul_ext_antisym:
  assumes "locally_antisymmetric ns (A + B)"
  and "locally_included ns s (A + B)"
  and "locally_irreflexive s (A + B)"
  and "locally_transitive s (A + B)"
  and "locally_transitive ns (A + B)"
  and "locally_compatible_l s ns (A + B)"
  and "locally_compatible_r ns s (A + B)"
  and "locally_splittable ns s (op =) (A + B)"
  shows "ns_mul_ext ns s A B \<longrightarrow> ns_mul_ext ns s B A \<longrightarrow> A = B"
proof-
  from assms and ns_mul_ext_antisym_aux[of _ _ _ _ "size A"] show ?thesis by blast
qed

text{* The orders are diff-compatible *}

lemma s_mul_ext_diff_single_compat:
  assumes AsB: "(A,B) \<in> (s_mul_ext ns s)"
  and "x \<in># A" and "x \<in># B"
  and "locally_irreflexive s A"           (is "?li A")
  and "locally_transitive s (A + B) (A + B) (A + B)"      (is "?lt_s (A + B)")
  and "locally_transitive ns (A + B) (A + B) (A + B)"     (is "?lt_ns (A + B)")
  and "locally_compatible_l ns s (A + B)" (is "?lcl (A + B)")
  and "locally_compatible_r ns s (A + B)" (is "?lcr (A + B)")
  shows "((A - {#x#}),(B - {#x#})) \<in> (s_mul_ext ns s)"
proof-
  from AsB obtain as bs A' B' where
    A_part: "A = multiset_of as + A'" and
    B_part: "B = multiset_of bs + B'" and
    aslbs: "length as = length bs" and
    A'_ne: "A' \<noteq> {#}" and
    ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> ((as ! i),(bs ! i)) \<in> ns" and
    s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
  {
    assume x_in_A': "x \<in># A'"
     and x_in_B': "x \<in># B'"
    from x_in_B' and s_A'_B' obtain a where
      a_in_A': "a \<in># A'" and
      s_a_x: "(a,x) \<in> s" by auto
    from a_in_A' and A_part have a_in_A: "a \<in># A" by simp
    with `?li A`[unfolded locally_irreflexive_def, rule_format]
     and s_a_x have "a \<noteq> x" by auto
    with a_in_A' have a_in_A'x: "a \<in># A' - {#x#}" by auto
    {
      fix b
      assume b_in_B': "b \<in># B'"
      with s_A'_B' obtain ab where
        ab_in_A': "ab \<in># A'" and
        s_ab_b: "(ab,b) \<in> s" by auto
      have "\<exists>a. a \<in># (A' - {#x#}) \<and> (a,b) \<in> s"
      proof (cases "ab = x")
       case True
        with b_in_B' and a_in_A' and ab_in_A' and s_ab_b and s_a_x
         and A_part and B_part and
         `?lt_s (A + B)`[unfolded locally_transitive_def, rule_format]
        have "(a,b) \<in> s" by auto
        with a_in_A'x show ?thesis by fast
       next
       case False
        with ab_in_A' have "ab \<in># (A' - {#x#})" by simp
        with s_ab_b show ?thesis by fast
      qed
    }
    hence s_A'x_B'x: "\<forall>b. b \<in># (B' - {#x#}) \<longrightarrow> (\<exists>a. a \<in># (A' - {#x#}) \<and> (a,b) \<in> s)" by simp
    from diff_union_single_conv and union_commute and A_part and x_in_A'
     have A_part2: "A - {#x#} = multiset_of as + (A' - {#x#})" by metis
    from diff_union_single_conv and union_commute and B_part and x_in_B'
     have B_part2: "B - {#x#} = multiset_of bs + (B' - {#x#})" by metis
    have "A' - {#x#} \<noteq> {#}"
     proof-
      from x_in_B' and s_A'_B' obtain a where
        a_in_A': "a \<in># A'" and
        s_a_x: "(a,x) \<in> s" by auto
      from a_in_A' and A_part have a_in_A: "a \<in># A" by simp
       with `?li A`[unfolded locally_irreflexive_def, rule_format]
       and s_a_x have "a \<noteq> x" by auto
      with a_in_A' have "a \<in># (A' - {#x#})" by simp
      thus ?thesis by force
     qed
    with s_A'x_B'x and A_part2 and B_part2 and aslbs and ns_as_bs
     have ?thesis unfolding s_mul_ext_def by blast
  }
  moreover
(*  hence CA'B': "x \<in># A' \<Longrightarrow> x \<in># B' \<Longrightarrow> ?thesis" by assumption *)
(** And this was supposed to be the easy case... ¬_¬ *)
  {
    assume x_in_as: "x \<in># multiset_of as"
     and x_in_B': "x \<in># B'"
    from x_in_B' and s_A'_B' obtain a where
      a_in_A': "a \<in># A'" and
      s_a_x: "(a,x) \<in> s" by auto
    from in_multiset_of_idx[OF x_in_as] obtain i where
      i_bd: "i < length as" and
      eq_x_as: "x = as ! i" by fast
    with ns_as_bs and aslbs have nsab: "((as ! i),(bs ! i)) \<in> ns" by force
    from x_in_as and A_part have x_in_A: "x \<in># A" by simp
    from a_in_A' and A_part have a_in_A: "a \<in># A" by simp
    from i_bd and aslbs and nth_mem_multiset_of have "bs ! i \<in># multiset_of bs" by auto
    with B_part have "bs ! i \<in># B" by simp
    with a_in_A and x_in_A and eq_x_as and s_a_x and nsab
     `?lcr (A + B)`[unfolded locally_compatible_r_def, rule_format]
    have s_a_bs: "(a,bs ! i) \<in> s" by auto
    with s_A'_B' and a_in_A'
    have s_A'_B'x: "\<forall>b. b \<in># (B' + {#bs ! i#} - {#x#}) \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> s)" by auto
    from remove_nth_soundness and i_bd
     have as': "multiset_of (remove_nth i as) = multiset_of as - {#as ! i#}" by simp
    from remove_nth_soundness and i_bd and aslbs
     have bs': "multiset_of (remove_nth i bs) = multiset_of bs - {#bs ! i#}" by simp
    from i_bd and nth_mem_multiset_of have "as ! i \<in># multiset_of as" by auto
    with eq_x_as and as' and A_part and diff_union_swap_multiset
     have A_part2: "A - {#x#} = multiset_of (remove_nth i as) + A'" by metis
    from i_bd and aslbs and nth_mem_multiset_of have b_in_bs: "bs ! i \<in># multiset_of bs"
     by auto
    with B_part and bs' and insert_DiffM2[OF b_in_bs] and diff_union_single_conv[OF x_in_B']
     have Bp: "B - {#x#} = multiset_of (remove_nth i bs) + {#bs ! i#} + (B' - {#x#})"
     by metis
   from diff_union_swap_multiset and union_commute and x_in_B' 
    have "{#bs ! i#} + (B' - {#x#}) = B' + {#bs ! i#} - {#x#}" by metis
   with Bp and union_assoc
    have B_part2: "B - {#x#} = multiset_of (remove_nth i bs) + (B' + {#bs ! i#} - {#x#})"
    by metis
  from i_bd and aslbs and remove_nth_length
   have aislbis: "length (remove_nth i as) = length (remove_nth i bs)" by auto
  note ns_ais_bis = remove_nth_P_compat[OF aslbs[symmetric] ns_as_bs, of i]
  from A_part2 and B_part2 and aislbis and A'_ne and s_A'_B'x and ns_ais_bis
   have ?thesis unfolding s_mul_ext_def by blast
  }
  moreover
(*  hence CasB': "x \<in># multiset_of as \<Longrightarrow> x \<in># B' \<Longrightarrow> ?thesis" by assumption*)
  {
    assume x_in_A': "x \<in># A'"
     and x_in_bs: "x \<in># multiset_of bs"
    with B_part have x_in_B: "x \<in># B" by simp
    from in_multiset_of_idx[OF x_in_bs] obtain i where
      i_bd: "i < length bs" and
      eq_x_bs: "x = bs ! i" by auto
    with ns_as_bs have ns_as_x: "((as ! i),x) \<in> ns" by simp
    have s_A'asx_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># (A' + {#as ! i#} - {#x#}) \<and> (a,b) \<in> s)"
    proof-
    {
      fix b assume b_in_B': "b \<in># B'"
      with s_A'_B' obtain a where
        a_in_A': "a \<in># A'" and
        s_a_b: "(a,b) \<in> s" by auto
      from b_in_B' and B_part have b_in_B: "b \<in># B" by simp
      hence "\<exists>a. a \<in># (A' + {#as ! i#} - {#x#}) \<and> (a,b) \<in> s" proof (cases "x = a")
        case True note eq_x_a = this
         thus ?thesis proof (cases "x = as ! i")
           case True
             hence "A' + {#as ! i#} - {#x#} = A'" by auto
             with a_in_A' and s_a_b show ?thesis by auto
           next
           case False
             hence as_in_A'asx: "(as ! i) \<in># (A' + {#as ! i#} - {#x#})" by simp
             from aslbs and i_bd have "i < length as" by simp
             with A_part and nth_mem_multiset_of have as_in_A: "as ! i \<in># A" by auto
             from s_a_b and ns_as_x and eq_x_a and b_in_B and x_in_B and as_in_A and
              `?lcl (A + B)`[unfolded locally_compatible_l_def, rule_format]
             have s_as_b: "((as ! i),b) \<in> s" by auto
             with False show ?thesis by auto
          qed
       next
       case False
         with a_in_A' and s_a_b show ?thesis by auto
     qed     
    }
    thus ?thesis by auto
    qed
    from remove_nth_soundness and i_bd
     have bs': "multiset_of (remove_nth i bs) = multiset_of bs - {#bs ! i#}" by simp
    from remove_nth_soundness and i_bd and aslbs[symmetric]
     have as': "multiset_of (remove_nth i as) = multiset_of as - {#as ! i#}" by simp
    from i_bd and nth_mem_multiset_of have "bs ! i \<in># multiset_of bs" by auto
    with eq_x_bs[symmetric] and bs' and B_part and diff_union_swap_multiset
     have B_part2: "B - {#x#} = multiset_of (remove_nth i bs) + B'" by metis
    from i_bd and aslbs[symmetric] and nth_mem_multiset_of
     have a_in_as: "as ! i \<in># multiset_of as"
     by auto
    with A_part and as' and insert_DiffM2[OF a_in_as] and diff_union_single_conv[OF x_in_A']
     have Ap: "A - {#x#} = multiset_of (remove_nth i as) + {#as ! i#} + (A' - {#x#})"
     by metis
   from diff_union_swap_multiset and union_commute and x_in_A'
    have "{#as ! i#} + (A' - {#x#}) = A' + {#as ! i#} - {#x#}" by metis
   with Ap and union_assoc
    have A_part2: "A - {#x#} = multiset_of (remove_nth i as) + (A' + {#as ! i#} - {#x#})"
    by metis
  from i_bd and aslbs and remove_nth_length
   have aislbis: "length (remove_nth i as) = length (remove_nth i bs)" by auto
  note ns_ais_bis = remove_nth_P_compat[OF aslbs[symmetric] ns_as_bs, of i]
  from A'_ne have A'asx_ne: "A' + {#as ! i#} - {#x#} \<noteq> {#}"
    proof (cases "x = as ! i")
      case True
       hence "A' + {#as ! i#} - {#x#} = A'" by simp
       with A'_ne show ?thesis by simp
      next
      case False
       from diff_union_swap[OF this]
        have "A' + {#as ! i#} - {#x#} = (A' - {#x#}) + {#as ! i#}" by simp
       thus ?thesis by auto
   qed
  from A_part2 and B_part2 and aislbis and A'asx_ne and s_A'asx_B' and ns_ais_bis
   have ?thesis unfolding s_mul_ext_def by blast
  }
  moreover
  {
    assume x_in_as: "x \<in># multiset_of as"
     and x_in_bs: "x \<in># multiset_of bs"
    with A_part have x_in_A: "x \<in># A" by simp
    from in_multiset_of_idx[OF x_in_as] obtain ia where
      ia_bd: "ia < length as" and
      eq_x_as: "x = as ! ia" by auto
    from in_multiset_of_idx[OF x_in_bs] obtain ib where
      ib_bd: "ib < length bs" and
      eq_x_bs: "x = bs ! ib" by auto
    have ?thesis proof (cases "ia = ib")
      case True note eq_ia_ib = this
        from nth_mem_multiset_of[OF ia_bd] and eq_x_as
         have x_in_as: "x \<in># multiset_of as" by simp
        from remove_nth_soundness[OF ia_bd] and eq_x_as
         and diff_union_swap_multiset[OF x_in_as] and A_part
         have A_part2: "A -{#x#} = multiset_of (remove_nth ia as) + A'" by auto
        from nth_mem_multiset_of[OF ib_bd] and eq_x_bs
         have x_in_bs: "x \<in># multiset_of bs" by simp
        from remove_nth_soundness[OF ib_bd] and eq_x_bs and eq_ia_ib
         and diff_union_swap_multiset[OF x_in_bs] and B_part
         have B_part2: "B -{#x#} = multiset_of (remove_nth ia bs) + B'" by auto
        from remove_nth_length[OF ia_bd] and remove_nth_length[OF ib_bd]
         and eq_ia_ib and aslbs
         have "length (remove_nth ia as) = length (remove_nth ia bs)" by auto
        with remove_nth_P_compat[OF aslbs[symmetric] ns_as_bs, of ia] and A_part2 and B_part2
         and A'_ne and s_A'_B' show ?thesis unfolding s_mul_ext_def by blast
      next
      case False note neq_ia_ib = this
       show ?thesis proof (cases "ia < ib")
        case True
        {
          fix ja jb
          assume ja_lt_jb: "ja < jb"
           and ja_bd: "ja < length as"
           and jb_bd: "jb < length bs"
           and eq_x_ja: "x = as ! ja"
           and eq_x_jb: "x = bs ! jb"
          from jb_bd and aslbs have jb_bd2: "jb < length as" by simp
          from ja_lt_jb and ja_bd and remove_nth_length[OF jb_bd2]
           have ja_bd2: "ja < length (remove_nth jb as)" by simp
          note as_in_as = nth_mem_multiset_of[OF jb_bd2]
(* Proving that A - x = {as ! jb + as - as ! jb - as ! ja} + A' *)
          from remove_nth_sound_l[OF ja_lt_jb] and remove_nth_soundness[OF jb_bd2]
           and nth_mem_multiset_of[OF ja_bd2]
           have as_in_asjb: "as ! ja \<in># multiset_of as - {#as ! jb#}" by metis
          from remove_nth_soundness[OF jb_bd2] and remove_nth_soundness[OF ja_bd2]
           and remove_nth_sound_l[OF ja_lt_jb]
           have "multiset_of (remove_nth ja (remove_nth jb as))
             = multiset_of as - {#as ! jb#} - {#as ! ja#}" by metis
           with diff_union_swap_single[OF as_in_asjb, of "as ! jb"]
            have "multiset_of ((as ! jb) # (remove_nth ja (remove_nth jb as))) =
              multiset_of as - {#as ! jb#} + {#as ! jb#} - {#as ! ja#}" by simp
           with diff_union_swap_single[OF as_in_as, of "as ! jb"]
            have "multiset_of ((as ! jb) # (remove_nth ja (remove_nth jb as))) =
             multiset_of as + {#as ! jb#} - {#as ! jb#} - {#as ! ja#}" by simp
           with insert_DiffM2[OF as_in_as]
            have "multiset_of ((as ! jb) # (remove_nth ja (remove_nth jb as))) =
             multiset_of as - {#as ! ja#}" by simp
           with A_part and eq_x_ja and nth_mem_multiset_of[OF ja_bd]
            and diff_union_swap_multiset
            have A_part2:
              "A - {#x#} = multiset_of ((as ! jb) # (remove_nth ja (remove_nth jb as))) + A'"
            by metis
(* Proving that B - x = {bs ! ja + bs - bs ! jb - bs ! ja} + B' *)
          from ja_lt_jb and ja_bd and aslbs and remove_nth_length[OF jb_bd2]
           have ja_bd2: "ja < length (remove_nth jb bs)" by simp
          note bs_in_bs = nth_mem_multiset_of[OF jb_bd]
          from remove_nth_sound_l[OF ja_lt_jb] and remove_nth_soundness[OF jb_bd]
           and nth_mem_multiset_of[OF ja_bd2]
           have bs_in_bsjb: "bs ! ja \<in># multiset_of bs - {#bs ! jb#}" by metis
          from remove_nth_soundness[OF jb_bd] and remove_nth_soundness[OF ja_bd2]
           and remove_nth_sound_l[OF ja_lt_jb]
           have "multiset_of (remove_nth ja (remove_nth jb bs))
             = multiset_of bs - {#bs ! jb#} - {#bs ! ja#}" by metis
           with insert_DiffM2[OF bs_in_bsjb]
            have "multiset_of ((bs ! ja) # (remove_nth ja (remove_nth jb bs))) =
             multiset_of bs - {#bs ! jb#}" by simp
           with B_part and eq_x_jb and nth_mem_multiset_of[OF jb_bd]
            and diff_union_swap_multiset
            have B_part2:
              "B - {#x#} = multiset_of ((bs ! ja) # (remove_nth ja (remove_nth jb bs))) + B'"
            by metis
(* Proving that the lists have the same length *)
          from aslbs and remove_nth_length have aislbis:
            "length ((as ! jb) # (remove_nth ja (remove_nth jb as)))
            = length ((bs ! ja) # (remove_nth ja (remove_nth jb bs)))"
            by auto
(* Proving that ns ({as ! jb + as - as ! jb - as ! ja} ! i) ({bs ! ja + bs - bs ! jb - bs ! ja} ! i) *)
          note ns_tmp = remove_nth_P_compat[OF aslbs[symmetric] ns_as_bs, of jb]
          from remove_nth_length and aslbs and jb_bd
           have "length (remove_nth jb as) = length (remove_nth jb bs)" by auto
          note ns_pre = remove_nth_P_compat[OF this[symmetric] ns_tmp, of ja]
          from ns_as_bs and jb_bd have ns1: "((as ! jb),(bs ! jb)) \<in> ns" by simp
          from ns_as_bs and ja_bd and aslbs have ns2: "((as ! ja),(bs ! ja)) \<in> ns" by simp
          from jb_bd and aslbs and  nth_mem_multiset_of
           have "as ! jb \<in># multiset_of as" by metis
          with A_part have asjb_in_A: "as ! jb \<in># A" by simp
          from ja_bd and aslbs and nth_mem_multiset_of
           have "bs ! ja \<in># multiset_of bs" by metis
          with B_part have bsja_in_B: "bs ! ja \<in># B" by simp
          from ns1 and ns2 and eq_x_ja and eq_x_jb and nth_mem_multiset_of[OF ja_bd]
           and nth_mem_multiset_of[OF jb_bd] and x_in_A and asjb_in_A and bsja_in_B and
           `?lt_ns (A + B)`[unfolded locally_transitive_def, rule_format]
          have ns_h: "((as ! jb),(bs ! ja)) \<in> ns" by auto
          {
            fix i
            assume i_bd: "i < length ((bs ! ja) # (remove_nth ja (remove_nth jb bs)))"
            have "  ((((as ! jb) # (remove_nth ja (remove_nth jb as))) ! i),
            (((bs ! ja) # (remove_nth ja (remove_nth jb bs))) ! i)) \<in> ns"
            proof (cases i)
              case 0
               with ns_h show ?thesis by simp
              next
              case (Suc j)
               with i_bd have j_bd: "j < length (remove_nth ja (remove_nth jb bs))" by simp
               from Suc and j_bd and ns_pre show ?thesis by simp
            qed
          }
          hence ns_ais_bis: "\<forall>i. i < length ((bs ! ja) # (remove_nth ja (remove_nth jb bs))) \<longrightarrow>
            ((((as ! jb) # (remove_nth ja (remove_nth jb as))) ! i),
            (((bs ! ja) # (remove_nth ja (remove_nth jb bs))) ! i)) \<in> ns" by simp
          from A_part2 and B_part2 and aislbis and A'_ne and s_A'_B' and ns_ais_bis
           have ?thesis unfolding s_mul_ext_def by blast
       }
       with True and ia_bd and ib_bd and eq_x_as and eq_x_bs show ?thesis by metis
       next
       case False
       with neq_ia_ib have ib_lt_ia: "ib < ia" by arith
        {
          fix ja jb
          assume jb_lt_ja: "jb < ja"
           and ja_bd: "ja < length as"
           and jb_bd: "jb < length bs"
           and eq_x_ja: "x = as ! ja"
           and eq_x_jb: "x = bs ! jb"
          from ja_bd and aslbs have ja_bd2: "ja < length bs" by simp
          from jb_lt_ja and jb_bd and remove_nth_length[OF ja_bd2]
           have jb_bd2: "jb < length (remove_nth ja bs)" by simp
          note bs_in_bs = nth_mem_multiset_of[OF ja_bd2]
          from remove_nth_sound_l[OF jb_lt_ja] and remove_nth_soundness[OF ja_bd2]
           and nth_mem_multiset_of[OF jb_bd2]
           have bs_in_bsja: "bs ! jb \<in># multiset_of bs - {#bs ! ja#}" by metis
          from remove_nth_soundness[OF ja_bd2] and remove_nth_soundness[OF jb_bd2]
           and remove_nth_sound_l[OF jb_lt_ja]
           have "multiset_of (remove_nth jb (remove_nth ja bs))
             = multiset_of bs - {#bs ! ja#} - {#bs ! jb#}" by metis
           with diff_union_swap_single[OF bs_in_bsja, of "bs ! ja"]
            have "multiset_of ((bs ! ja) # (remove_nth jb (remove_nth ja bs))) =
              multiset_of bs - {#bs ! ja#} + {#bs ! ja#} - {#bs ! jb#}" by simp
           with diff_union_swap_single[OF bs_in_bs, of "bs ! ja"]
            have "multiset_of ((bs ! ja) # (remove_nth jb (remove_nth ja bs))) =
             multiset_of bs + {#bs ! ja#} - {#bs ! ja#} - {#bs ! jb#}" by simp
           with insert_DiffM2[OF bs_in_bs]
            have "multiset_of ((bs ! ja) # (remove_nth jb (remove_nth ja bs))) =
             multiset_of bs - {#bs ! jb#}" by simp
           with B_part and eq_x_jb and nth_mem_multiset_of[OF jb_bd]
            and diff_union_swap_multiset
            have B_part2: 
              "B - {#x#} = multiset_of ((bs ! ja) # (remove_nth jb (remove_nth ja bs))) + B'"
            by metis
          from jb_lt_ja and jb_bd and aslbs and remove_nth_length[OF ja_bd2]
           have jb_bd2: "jb < length (remove_nth ja as)" by simp
          note as_in_as = nth_mem_multiset_of[OF ja_bd]
          from remove_nth_sound_l[OF jb_lt_ja] and remove_nth_soundness[OF ja_bd]
           and nth_mem_multiset_of[OF jb_bd2]
           have as_in_asja: "as ! jb \<in># multiset_of as - {#as ! ja#}" by metis
          from remove_nth_soundness[OF ja_bd] and remove_nth_soundness[OF jb_bd2]
           and remove_nth_sound_l[OF jb_lt_ja]
           have "multiset_of (remove_nth jb (remove_nth ja as))
             = multiset_of as - {#as ! ja#} - {#as ! jb#}" by metis
           with insert_DiffM2[OF as_in_asja]
            have "multiset_of ((as ! jb) # (remove_nth jb (remove_nth ja as))) =
             multiset_of as - {#as ! ja#}" by simp
           with A_part and eq_x_ja and nth_mem_multiset_of[OF ja_bd]
            and diff_union_swap_multiset
            have A_part2:
              "A - {#x#} = multiset_of ((as ! jb) # (remove_nth jb (remove_nth ja as))) + A'"
            by metis
(* Proving that the lists have the same length *)
          from aslbs and remove_nth_length have aislbis:
            "length ((bs ! ja) # (remove_nth jb (remove_nth ja bs)))
            = length ((as ! jb) # (remove_nth jb (remove_nth ja as)))"
            by auto
(* Proving that ns ({as ! jb + as - as ! jb - as ! ja} ! i) ({bs ! ja + bs - bs ! jb - bs ! ja} ! i) *)
          note ns_tmp = remove_nth_P_compat[OF aslbs[symmetric] ns_as_bs, of ja]
          from remove_nth_length and aslbs and ja_bd
           have "length (remove_nth ja bs) = length (remove_nth ja as)" by auto
          note ns_pre = remove_nth_P_compat[OF this ns_tmp, of jb]
          from ns_as_bs and ja_bd and aslbs have ns1: "((as ! ja),(bs ! ja)) \<in> ns" by simp
          from ns_as_bs and jb_bd have ns2: "((as ! jb),(bs ! jb)) \<in> ns" by simp
          from ja_bd and aslbs and nth_mem_multiset_of
           have "bs ! ja \<in># multiset_of bs" by metis
          with B_part have bsja_in_B: "bs ! ja \<in># B" by simp
          from jb_bd and aslbs and nth_mem_multiset_of
           have "as ! jb \<in># multiset_of as" by metis
          with A_part have asjb_in_A: "as ! jb \<in># A" by simp
          from ns1 and ns2 and eq_x_ja and eq_x_jb and nth_mem_multiset_of[OF jb_bd]
           and nth_mem_multiset_of[OF ja_bd] and x_in_A and asjb_in_A and bsja_in_B and
           `?lt_ns (A + B)`[unfolded locally_transitive_def, rule_format]
          have ns_h: "((as ! jb),(bs ! ja)) \<in> ns" by auto
          {
            fix i
            assume i_bd: "i < length ((bs ! ja) # (remove_nth jb (remove_nth ja bs)))"
            have "((((as ! jb) # (remove_nth jb (remove_nth ja as))) ! i),
            (((bs ! ja) # (remove_nth jb (remove_nth ja bs))) ! i)) \<in> ns"
            proof (cases i)
              case 0
               with ns_h show ?thesis by simp
              next
              case (Suc j)
               with i_bd have j_bd: "j < length (remove_nth jb (remove_nth ja bs))" by simp
               from Suc and j_bd and ns_pre show ?thesis by simp
            qed
          }
          hence ns_ais_bis: "\<forall>i. i < length ((bs ! ja) # (remove_nth jb (remove_nth ja bs))) \<longrightarrow>
            ((((as ! jb) # (remove_nth jb (remove_nth ja as))) ! i),
            (((bs ! ja) # (remove_nth jb (remove_nth ja bs))) ! i)) \<in> ns" by simp
          from A_part2 and B_part2 and aislbis[symmetric] and A'_ne
           and s_A'_B' and ns_ais_bis
           have ?thesis unfolding s_mul_ext_def by blast
       }
       with ib_lt_ia and ia_bd and ib_bd and eq_x_as and eq_x_bs show ?thesis by metis
  qed
  qed
  }
  ultimately show ?thesis using A_part and B_part and `x \<in># A` and `x \<in># B` by auto
qed


lemma s_mul_ext_diff_multiset_compat:
  assumes "(A,B) \<in> (s_mul_ext ns s)"
  and "C \<le> A" and "C \<le> B"
  and "locally_irreflexive s A"           (is "?li A")
  and "locally_transitive s (A + B)"      (is "?lt_s (A + B)")
  and "locally_transitive ns (A + B)"     (is "?lt_ns (A + B)")
  and "locally_compatible_l ns s (A + B)" (is "?lcl (A + B)")
  and "locally_compatible_r ns s (A + B)" (is "?lcr (A + B)")
  shows "((A - C),(B - C)) \<in> (s_mul_ext ns s)"
using assms
proof (induct C arbitrary: A B)
 case empty
  thus ?case by simp
 next
 case (add C' x) note IH = this
  have x_in_C: "x \<in># (C' + {#x#})" by simp
  from x_in_C and `C' + {#x#} \<le> A` and mset_leD have x_in_A: "x \<in># A" by metis
  from x_in_C and `C' + {#x#} \<le> B` and mset_leD have x_in_B: "x \<in># B" by metis
  from x_in_A and x_in_B and IH and s_mul_ext_diff_single_compat
   have sAx_Bx: "((A - {#x#}),(B - {#x#})) \<in> (s_mul_ext ns s)"
   by metis
  from x_in_A and x_in_B and diff_add and diff_union_single_conv and union_commute
   have Hrew: "A + B - {#x#} - {#x#} = (A - {#x#}) + (B - {#x#})" by metis
  from IH(3) and diff_union_cancelR and diff_union_swap_multiset
   and mset_le_mono_add_right_cancel and x_in_A
   have C'_subA: "C' \<le> A - {#x#}" by metis
  from IH(4) and diff_union_cancelR and diff_union_swap_multiset
   and mset_le_mono_add_right_cancel and x_in_B
   have C'_subB: "C' \<le> B - {#x#}" by metis
  from IH(5) and li_minus have li': "?li (A - {#x#})" by auto
  from IH(6) and lt_minus have "?lt_s (A + B - {#x#} - {#x#})" by metis
  with Hrew have lt_s': "?lt_s ((A - {#x#}) + (B - {#x#}))" by simp
  from IH(7) and lt_minus have "?lt_ns (A + B - {#x#} - {#x#})" by metis
  with Hrew have lt_ns': "?lt_ns ((A - {#x#}) + (B - {#x#}))" by simp
  from IH(8) and lcl_minus have "?lcl (A + B - {#x#} - {#x#})" by metis
  with Hrew have lcl': "?lcl ((A - {#x#}) + (B - {#x#}))" by simp
  from IH(9) and lcr_minus have "?lcr (A + B - {#x#} - {#x#})" by metis
  with Hrew have lcr': "?lcr ((A - {#x#}) + (B - {#x#}))" by simp
  from IH and sAx_Bx and C'_subA and C'_subB and li' and lt_s' and lt_ns'
   and lcl' and lcr' have "((A - {#x#} - C'),(B - {#x#} - C')) \<in> (s_mul_ext ns s)"
   by simp
  with diff_add and union_commute show ?case by metis
qed  



text{* The strict order is irreflexive. *}

lemma s_mul_ext_irrefl_aux:
  assumes "locally_irreflexive s A"      (is "?li A")
  and "locally_transitive s A A A"       (is "?lt_s A")
  and "locally_transitive ns A A A"      (is "?lt_ns A")
  and "locally_compatible_l ns s A A A"  (is "?lcl A")
  and "locally_compatible_r ns s A A A"  (is "?lcr A")
  and "size A = n"                       (is "?sA A n")
  shows "(A,A) \<notin> (s_mul_ext ns s)"
using assms
proof (induct n arbitrary: A)
 case 0
  thus ?case by simp
 next
 case (Suc m)
  with size_eq_Suc_imp_eq_union obtain a A' where
    A_part: "A = A' + {#a#}" by metis
  hence A'_part: "A - {#a#} = A'" by auto
  from A_part have a_in_A: "a \<in># A" by simp
  from Suc(2) and A'_part and li_minus have li': "?li A'" by auto
  from Suc(3) and A_part and lt_trans_l have lt_s': "?lt_s A'" by auto
  from Suc(4) and A_part and lt_trans_l have lt_ns': "?lt_ns A'" by auto
  from Suc(5) and A_part and lcl_trans_l have lcl': "?lcl A'" by auto
  from Suc(6) and A_part and lcr_trans_l have lcr': "?lcr A'" by metis
  from Suc(7) and A_part have sA': "size A' = m " by simp
  from Suc(1) and li' and lt_s' and lt_ns' and lcl' and lcr' and sA'
  have nA'sA': "(A',A') \<notin> (s_mul_ext ns s)" by simp
  have x_in_AA: "x \<in># (A + A) \<Longrightarrow> x \<in># A" by auto
  {
    fix t u v
    assume "t \<in># (A + A)" "u \<in># (A + A)" "v \<in># (A + A)"
    "(t,u) \<in> s" "(u,v) \<in> s"
    with x_in_AA have "t \<in># A" and "u \<in># A" and "v \<in># A" by auto
    with Suc(3) and `(t,u) \<in> s` `(u,v) \<in> s` have "(t,v) \<in> s"
     unfolding locally_transitive_def by metis
  }
  hence lt_s': "?lt_s (A + A)" unfolding locally_transitive_def by metis
  {
    fix t u v
    assume "t \<in># (A + A)" "u \<in># (A + A)" "v \<in># (A + A)"
    "(t,u) \<in> ns" "(u,v) \<in> ns"
    with x_in_AA have "t \<in># A" and "u \<in># A" and "v \<in># A" by auto
    with Suc(4) and `(t,u) \<in> ns` `(u,v) \<in> ns` have "(t,v) \<in> ns"
     unfolding locally_transitive_def by metis
  }
  hence lt_ns': "?lt_ns (A + A)" unfolding locally_transitive_def by metis
  {
    fix t u v
    assume "t \<in># (A + A)" "u \<in># (A + A)" "v \<in># (A + A)"
    "(t,u) \<in> s" "(u,v) \<in> ns"
    with x_in_AA have "t \<in># A" and "u \<in># A" and "v \<in># A" by auto
    with Suc(6) and `(t,u) \<in> s` `(u,v) \<in> ns` have "(t,v) \<in> s"
     unfolding locally_compatible_r_def by metis
  }
  hence lcr': "?lcr (A + A)" unfolding locally_compatible_r_def by metis
  {
    fix t u v
    assume "t \<in># (A + A)" "u \<in># (A + A)" "v \<in># (A + A)"
    "(t,u) \<in> ns" "(u,v) \<in> s"
    with x_in_AA have "t \<in># A" and "u \<in># A" and "v \<in># A" by auto
    with Suc(5) and `(t,u) \<in> ns` `(u,v) \<in> s` have "(t,v) \<in> s"
     unfolding locally_compatible_l_def by metis
  }
  hence lcl': "?lcl (A + A)" unfolding locally_compatible_l_def by metis
  with A'_part and a_in_A and `?li A` and lt_s' and lt_ns' and lcl' and lcr'
   and s_mul_ext_diff_single_compat
   have "(A,A) \<in> (s_mul_ext ns s) \<Longrightarrow> (A',A') \<in> (s_mul_ext ns s)"
   by metis
  with nA'sA' show ?case by fast
qed

lemma s_mul_ext_irrefl:
  assumes "locally_irreflexive s A"
  and "locally_transitive s A"
  and "locally_transitive ns A"
  and "locally_compatible_l ns s A"
  and "locally_compatible_r ns s A"
  shows "(A,A) \<notin> (s_mul_ext ns s)"
using assms and s_mul_ext_irrefl_aux[of _ _ _ "size A"]
by metis


lemma multiset3_single3_diff :
  assumes "a \<in># A" and "b \<in># B" and "c \<in># C"
  shows "(A - {#a#}) + (B - {#b#}) + (C - {#c#})
  = (A + B + C) - ({#a#} + {#b#} + {#c#})"
proof-
  note AaBb = diff_union_single_conv[OF `b \<in># B`, of "A - {#a#}"]
  note AaBbCc = diff_union_single_conv[OF `c \<in># C`, of "A - {#a#} + (B - {#b#})"]
  from `a \<in># A` have "a \<in># (A + B)" by simp
  from diff_union_swap_multiset[OF `a \<in># A`] and diff_union_swap_multiset[OF this]
   have AaBC: "A + B + C - {#a#} = A - {#a#} + B + C" by simp
  from `b \<in># B` have "b \<in># (A - {#a#} + B)" by simp
  from diff_union_swap_multiset[OF this] and AaBb and AaBC
   have AaBbC: "A + B + C - {#a#} - {#b#} = A - {#a#} + (B - {#b#}) + C" by force
  with AaBbCc and diff_add show ?thesis by metis
qed
end
