theory Simple_Termination
imports
  "$AFP/Well_Quasi_Orders/Wpo_Kruskal_Terms"
  Signature_Extension
begin

interpretation "term": finite_tree Fun "fst \<circ> the \<circ> root" args
  by (unfold_locales) auto

abbreviation "emb \<equiv> term.term_hemb"
abbreviation "gterms \<equiv> term.terms"

inductive_set terms for F where
  "Var x \<in> terms F" |
  "(f, n) \<in> F \<Longrightarrow> length ts = n \<Longrightarrow> \<forall>t\<in>set ts. t \<in> terms F \<Longrightarrow> Fun f ts \<in> terms F"

lemma ground_funas_term_subseteq_gterms_conv:
  "ground t \<and> funas_term t \<subseteq> F \<longleftrightarrow> t \<in> gterms F"
proof
  assume "ground t \<and> funas_term t \<subseteq> F"
  then show "t \<in> gterms F"
    by (induct t) (auto, rule term.terms.intros, auto)
next
  assume "t \<in> gterms F"
  then show "ground t \<and> funas_term t \<subseteq> F"
    by (induct) auto
qed

lemma gtermsD: "t \<in> gterms F \<Longrightarrow> ground t \<and> funas_term t \<subseteq> F"
  by (metis ground_funas_term_subseteq_gterms_conv)

lemma terms_UNIV [simp]:
  "terms UNIV = (UNIV :: ('f, 'v) term set)"
proof
  show "(UNIV :: ('f, 'v) term set) \<subseteq> terms UNIV"
  proof
    fix t :: "('f, 'v) term"
    show "t \<in> terms UNIV"
      by (induct t) (auto intro: terms.intros)
  qed
qed auto

lemma wfp_on_gterms_with_const_imp_wfp_on:
  assumes "(c, 0) \<in> F"
  assumes "wfp_on P (gterms F)"
    and "\<And>s t \<sigma>. P s t \<Longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
  shows "wfp_on P (terms F)"
proof (rule ccontr)
  assume "\<not> wfp_on P (terms F)"
  then obtain f where *: "\<And>i. f i \<in> terms F"
    and **: "\<And>i. P (f (Suc i)) (f i)" unfolding wfp_on_def by blast
  let ?\<sigma> = "\<lambda>_. Fun c []"
  let ?f = "\<lambda>i. f i \<cdot> ?\<sigma>"
  from ** and assms have "\<And>i. P (?f (Suc i)) (?f i)" by auto
  moreover have "\<And>i. ?f i \<in> gterms F"
  proof -
    fix i
    from * have "f i \<in> terms F" by auto
    then show "?f i \<in> gterms F" using `(c, 0) \<in> F`
    apply (induct)
    apply force
    apply auto
    apply (rule term.terms.intros)
    apply assumption
    apply simp_all
    done
  qed
  ultimately have "\<forall>i. ?f i \<in> gterms F \<and> P (?f (Suc i)) (?f i)" by blast
  then have "\<exists>g. \<forall>i. g i \<in> gterms F \<and> P (g (Suc i)) (g i)" by auto
  then show False using assms unfolding wfp_on_def by blast
qed

definition rewrite_relation where
  "rewrite_relation P = (
    (\<forall>s t C. P s t \<longrightarrow> P (C\<langle>s\<rangle>) (C\<langle>t\<rangle>)) \<and>
    (\<forall>s t \<sigma>. P s t \<longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"

definition rewrite_order where
  "rewrite_order P = (rewrite_relation P \<and> po_on P UNIV)"

definition simplification_order_on where
  "simplification_order_on P F = (rewrite_order P \<and>
    (\<exists>Q. wpo_on Q F \<and> (\<forall>s t. emb F Q s t \<longrightarrow> P s t)))"

lemma gterms_imp_terms:
  fixes F :: "('a \<times> nat) set"
  shows "gterms F \<subseteq> terms F"
proof
  fix t :: "('a, 'b) term"
  assume "t \<in> gterms F"
  then show "t \<in> terms F"
    by (induct) (rule terms.intros, auto)
qed

lemma rev_chain_on_transp_on_less:
  assumes "\<And>i. f i \<in> A" and "\<And>i. P (f (Suc i)) (f i)"
    and "transp_on P A" and "i < j"
  shows "P (f j) (f i)"
proof -
  let ?P = "P\<inverse>\<inverse>"
  from assms have "chain_on ?P f A" by auto
  moreover
  from transp_on_converse [OF `transp_on P A`]
    have "transp_on ?P A" .
  ultimately
  show ?thesis
    using chain_on_transp_on_less [OF _ _ `i < j`] by blast
qed

lemma simplification_order_on_imp_wfp_on_gterms:
  assumes "simplification_order_on P F"
  shows "wfp_on P (gterms F)"
proof (rule ccontr)
  from assms obtain Q
    where wpo: "wpo_on Q F"
    and po: "po_on P UNIV"
    and emb: "\<And>s t. emb F Q s t \<Longrightarrow> P s t"
    unfolding simplification_order_on_def rewrite_order_def
    by blast
  assume "\<not> wfp_on P (gterms F)"
  then obtain f where *: "\<And>i. f i \<in> gterms F"
    and **: "\<And>i. P (f (Suc i)) (f i)"
    unfolding wfp_on_def by blast
  from transp_on_subset [OF subset_UNIV po [THEN po_on_imp_transp_on]]
    have trans': "transp_on P (gterms F)" .
  from term.wpo_on_terms [OF wpo]
    have "wpo_on (emb F Q) (gterms F)" .
  with * obtain i j where "i < j" and "(emb F Q)\<^sup>=\<^sup>= (f i) (f j)"
    unfolding wpo_on_def almost_full_on_def good_def by blast
  with emb have "P\<^sup>=\<^sup>= (f i) (f j)" by auto
  moreover
  from rev_chain_on_transp_on_less [OF _ _ trans' `i < j`] and * and **
    have "P (f j) (f i)" by blast
  moreover
  from irreflp_on_subset [OF subset_UNIV po [THEN po_on_imp_irreflp_on]]
    have "irreflp_on P (gterms F)" .
  ultimately
    show False
    using po_on_imp_antisymp_on [OF po_on_subset [OF subset_UNIV po]] and *
    unfolding antisymp_on_def irreflp_on_def
    by (metis sup2CI)
qed

lemma funas_term_terms_conv:
  "funas_term t \<subseteq> F \<longleftrightarrow> t \<in> terms F"
proof
  assume "funas_term t \<subseteq> F"
  then show "t \<in> terms F" by (induct t) (auto intro: terms.intros)
next
  assume "t \<in> terms F"
  then show "funas_term t \<subseteq> F" by (induct) auto
qed

lemma wfp_on_terms_imp_SN_rstep:
  assumes subset: "funas_trs R \<subseteq> F"
    and compatible: "R \<subseteq> {(l, r). P r l}"
    and subst: "\<And>s t \<sigma>. P s t \<Longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
    and ctxt: "\<And>s t C. P s t \<Longrightarrow> P (C\<langle>s\<rangle>) (C\<langle>t\<rangle>)"
    and "wfp_on P (terms F)"
  shows "SN (rstep R)"
proof (rule ccontr)
  let ?F = "funas_trs R"
  let ?R = "sig_step ?F (rstep R)"
  assume "\<not> SN (rstep R)"
  then have "\<not> SN ?R" unfolding SN_sig_rstep_SN_rstep_conv .
  then obtain f where *: "\<And>i. (f i, f (Suc i)) \<in> ?R" by (auto simp: SN_defs)
  have "\<forall>i. f i \<in> terms F \<and> P (f (Suc i)) (f i)"
  proof
    fix i
    from * have "funas_term (f i) \<subseteq> F" 
      using subset unfolding sig_step_def by blast
    then have "f i \<in> terms F" by (simp add: funas_term_terms_conv)
    moreover
    have "P (f (Suc i)) (f i)"
    proof -
      from * have "(f i, f (Suc i)) \<in> ?R" by simp
      then have "(f i, f (Suc i)) \<in> rstep R" unfolding sig_step_def by blast
      then show ?thesis
      proof
        fix l r C \<sigma>
        assume "(l, r) \<in> R" and fi: "f i = C\<langle>l \<cdot> \<sigma>\<rangle>" and fsi: "f (Suc i) = C\<langle>r \<cdot> \<sigma>\<rangle>"
        from `(l, r) \<in> R` and compatible have "P r l" by auto
        from ctxt [OF subst [OF this]] show ?thesis unfolding fi fsi by blast
      qed
    qed
    ultimately show "f i \<in> terms F \<and> P (f (Suc i)) (f i)" by blast
  qed
  with `wfp_on P (terms F)` show False unfolding wfp_on_def by blast
qed

text {*Compatibility with a simplification-order implies strong normalization
of a TRS.*}
lemma simplification_order_imp_SN_rstep:
  assumes so: "simplification_order_on P F"
    and "(c, 0) \<in> F"
    and "funas_trs R \<subseteq> F"
    and "R \<subseteq> {(l, r). P r l}"
  shows "SN (rstep R)"
proof (rule wfp_on_terms_imp_SN_rstep)
  show "funas_trs R \<subseteq> F" by fact
  show "R \<subseteq> {(l, r). P r l}" by fact
  from so show "\<And>s t \<sigma>. P s t \<Longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
    and "\<And>s t C. P s t \<Longrightarrow> P C\<langle>s\<rangle> C\<langle>t\<rangle>"
    unfolding simplification_order_on_def rewrite_order_def rewrite_relation_def
    by auto
  show "wfp_on P (terms F)"
  proof (rule  wfp_on_gterms_with_const_imp_wfp_on)
    show "(c, 0) \<in> F" by fact
    show "\<And>s t \<sigma>. P s t \<Longrightarrow> P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)" by fact
    show "wfp_on P (gterms F)"
    proof (rule simplification_order_on_imp_wfp_on_gterms)
      show "simplification_order_on P F" by fact
    qed
  qed
qed

lemma wfp_on_SN_conv:
  "wfp_on P UNIV \<longleftrightarrow> SN {(x, y). P y x}"
  unfolding wfp_on_def SN_defs by blast

lemma SN_wfp_on_conv:
  "SN R \<longleftrightarrow> wfp_on (\<lambda>x y. (y, x) \<in> R) UNIV"
  unfolding wfp_on_def SN_defs by blast

(*
text {*Every well-founded partial-order can be extended to a well-partial-order.*}
lemma po_on_wfp_on_imp_wpo_on_extension:
  assumes "po_on P A" and "wfp_on P A"
  shows "\<exists>Q. (\<forall>x\<in>A. \<forall>y\<in>A. P x y \<longrightarrow> Q x y) \<and> wpo_on Q A"
  sorry
*)

end

